#ifndef CONSOLE_H
#define CONSOLE_H

#include <avr/eeprom.h>

#ifndef EEMEM
#define EEMEM  __attribute__ ((section (".eeprom")))
#endif

#define CONFIG_FLAG_DHCP     0x01   /* use DHCP */

#ifdef CONSOLE_CONFIG_WLAN
#define CONFIG_FLAG_MANAGED  0x02   /* use managed network */
#define CONFIG_KEY_WEP40     0x04   /* key sizes */
#define CONFIG_KEY_WEP104    0x08
#define CONFIG_MAX_SSID_SIZE 32
#define CONFIG_MAX_KEY_SIZE  13   // 104 bit == 13 bytes
#endif

/* this struct is being placed in eeprom */
typedef struct {
  /* shared part */
  unsigned char valid;
  unsigned char flags;

#ifdef CONSOLE_CONFIG_WLAN
  /* CONSOLE_CONFIG_WLAN part */
  unsigned char channel;
  unsigned char essid[CONFIG_MAX_SSID_SIZE+1];
  unsigned char key[CONFIG_MAX_KEY_SIZE];
#endif

  /* ip configuration */
  uip_ipaddr_t addr;
  uip_ipaddr_t mask;
  uip_ipaddr_t gate;
} config_t;

/* give other parts access to config stored in eeprom */
extern config_t config EEMEM;

void config_init(void);
void config_console_handle(void);
void config_console_fatal(const char *str, unsigned char n);
void config_eeprom_read_str(char *pointer_ram, char *pointer_eeprom);
unsigned char config_eeprom_strlen(char *pointer_eeprom);
void show_config(void);

#endif //CONSOLE_H
