/*
 * SPI2CF (c) 2006, Till Harbaum.
 * All rights reserved.
 *
 * This file is distributed under the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This file is part SPI2CF port of the uIP TCP/IP stack.
 */

/*
  prism2.c

  Based on the prism2.c file from pascal stangs avrlib.
*/

#include "uip.h"
#include "uip_arp.h"

#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include <avr/io.h>
#include <avr/pgmspace.h>

#include "cf.h"
#include "prism2.h"
#include "config.h"

// #define DEBUG_CMD
// #define DEBUG_EVENT
// #define DEBUG_BAP_READ
// #define DEBUG_BAP_WRITE

#define MAGIC 0x4711

// keep one global tx buffer
static uint16_t txfid;

#include <util/delay.h>

void prism2_enable(void) {
  uint8_t u;
  uint16_t s;

#if 0  
  printf_P(PSTR("\nPrism2 PCMCIA attribute registers:\n"));
  printf_P(PSTR("COR: 0x%02x\n"), cf_reg_read_byte(0x3e0 + PCMCIA_ATTR_COR));
  printf_P(PSTR("CSR: 0x%02x\n"), cf_reg_read_byte(0x3e0 + PCMCIA_ATTR_CSR));
  printf_P(PSTR("PRR: 0x%02x\n"), cf_reg_read_byte(0x3e0 + PCMCIA_ATTR_PRR));
  printf_P(PSTR("Enabling IO access ...\n"));
#endif

  cf_reg_write_byte(0x3e0 + PCMCIA_ATTR_COR, 
		    cf_reg_read_byte(0x3e0 + PCMCIA_ATTR_COR) | 0x41);

#if 0  
  printf_P(PSTR("COR: 0x%02x\n"), cf_reg_read_byte(0x3e0 + PCMCIA_ATTR_COR));
#endif

  // check for card availablity by writing and verifying one
  // of the software register
  cf_io_write_word(PRISM2_REG_SWSUP0, MAGIC);
  s = cf_io_read_word(PRISM2_REG_SWSUP0);
  if(s != MAGIC) {
    printf_P(PSTR("Register was written to %x, read back %x!\n"), s, MAGIC);
    config_console_fatal(PSTR("Test write failed"),2);
  }
} 

uint8_t prism2_command(uint16_t cmd, uint16_t param0) {
  uint16_t result;

  // wait until card not busy
#ifdef DEBUG_CMD
  printf_P(PSTR("PRISM_CMD: Wait until card ready\n"));
#endif
  while(cf_io_read_word(PRISM2_REG_CMD) & PRISM2_CMD_BUSY);
	
#ifdef DEBUG_CMD
  printf_P(PSTR("PRISM_CMD: Issue Command = 0x%x\n"), cmd);
#endif
  cf_io_write_word(PRISM2_REG_PARAM0, param0);
  cf_io_write_word(PRISM2_REG_CMD, cmd);
  
  // wait until card not busy
#ifdef DEBUG_CMD
  printf_P(PSTR("PRISM_CMD: Wait until card ready\n"));
#endif
  while(cf_io_read_word(PRISM2_REG_CMD) & PRISM2_CMD_BUSY);
  
  // read event register - wait for command to complete
#ifdef DEBUG_CMD
  printf_P(PSTR("PRISM_CMD: Wait for command to complete\n"));
#endif
  while(!(cf_io_read_word(PRISM2_REG_EVSTAT) & PRISM2_EVENT_CMD));
  
  // read status register
  result = cf_io_read_word(PRISM2_REG_STATUS);
#ifdef DEBUG_CMD
  printf_P(PSTR("PRISM_CMD: Result  = 0x%x\n"), result>>8);
  printf_P(PSTR("PRISM_CMD: CmdCode = 0x%x\n"), result);
#endif
  
  // acknowledge event
#ifdef DEBUG_CMD
  printf_P(PSTR("PRISM_CMD: Ack command event\n"));
#endif
  cf_io_write_word(PRISM2_REG_EVACK, PRISM2_EVENT_CMD);
  
  // return command result
  return result>>8;
}

uint8_t prism2_read_BAP0(uint16_t id, uint16_t offset, 
			 uint16_t* data, uint16_t len) {
  // wait for BAP to be ready
#ifdef DEBUG_BAP_READ
  printf_P(PSTR("PRISM_BAP: Wait1\n"));
#endif
  while( cf_io_read_word(PRISM2_REG_BAP0OFFSET) & PRISM2_BAPOFFSET_BUSY);
  // set ID
  cf_io_write_word(PRISM2_REG_BAP0SEL, id);
  // set offset
  cf_io_write_word(PRISM2_REG_BAP0OFFSET, offset);
  // wait for BAP to be ready
#ifdef DEBUG_BAP_READ
  printf_P(PSTR("PRISM_BAP: Wait2\n"));
#endif
  while( cf_io_read_word(PRISM2_REG_BAP0OFFSET) & PRISM2_BAPOFFSET_BUSY);
  // check for error
  if(cf_io_read_word(PRISM2_REG_BAP0OFFSET) & PRISM2_BAPOFFSET_ERROR)
    return -1;
  // read data
#ifdef DEBUG_BAP_READ
  printf_P(PSTR("PRISM_BAP: Read\n"));
#endif

  cf_io_read(PRISM2_REG_BAP0DATA, data, len);

  // return success
  return 0;
}

uint8_t prism2_write_BAP0(uint16_t id, uint16_t offset, 
			  uint16_t* data, uint16_t len) {

  // wait for BAP to be ready
#ifdef DEBUG_BAP_WRITE
  printf_P(PSTR("PRISM_BAP: Wait1\n"));
#endif
  while(cf_io_read_word(PRISM2_REG_BAP0OFFSET) & PRISM2_BAPOFFSET_BUSY);
  // set ID
  cf_io_write_word(PRISM2_REG_BAP0SEL, id);
  // set offset
  cf_io_write_word(PRISM2_REG_BAP0OFFSET, offset);
  // wait for BAP to be ready
#ifdef DEBUG_BAP_WRITE
  printf_P(PSTR("PRISM_BAP: Wait2\n"));
#endif
  while( cf_io_read_word(PRISM2_REG_BAP0OFFSET) & PRISM2_BAPOFFSET_BUSY);
  // check for error
  if(cf_io_read_word(PRISM2_REG_BAP0OFFSET) & PRISM2_BAPOFFSET_ERROR)
    return -1;
  // write data
#ifdef DEBUG_BAP_WRITE
  printf_P(PSTR("PRISM_BAP: Write\n"));
#endif
  cf_io_write(PRISM2_REG_BAP0DATA, data, len);

  // return success
  return 0;
}

uint8_t prism2_read_RID(uint16_t id, uint16_t offset, 
			uint16_t* data, uint16_t len) {
  prism2_command(PRISM2_CMD_ACCESS_RD, id);
  return prism2_read_BAP0(id, offset, data, len);
}

uint8_t prism2_write_RID(uint16_t offset, uint16_t* data) {
  uint8_t result;
  result = prism2_write_BAP0(data[1], offset, data, data[0]+1);
  prism2_command(PRISM2_CMD_ACCESS_WR, data[1]);
  return result;
}

uint8_t prism2_write_RID2(uint16_t id, uint16_t val) {
  uint16_t buffer[3] = { 2, id, val };
  uint8_t result;
  result = prism2_write_BAP0(id, 0, buffer, 3);
  prism2_command(PRISM2_CMD_ACCESS_WR, id);
  return result;
}

void prism2_set_SSID(void) {
  uint16_t buffer[(config_eeprom_strlen(config.essid)+7)/2];

  // prepare buffer for SSID write
  buffer[1] = (eeprom_read_byte(&config.flags) & CONFIG_FLAG_MANAGED)?
                PRISM2_RID_CNFDESIREDSSID : PRISM2_RID_CNFOWNSSID;
  buffer[2] = config_eeprom_strlen(config.essid);
  buffer[0] = (buffer[2]+5)/2;

  // copy ssid string to buffer, use read_block since we don't want the 0 byte
  eeprom_read_block((uint8_t*)&buffer[3], config.essid, buffer[2]);
  // write SSID
//  hexdump(buffer, 2 + 2*buffer[0]);
  prism2_write_RID(0, buffer);
}

void prism2_get_MAC(uint8_t *mac) {
  uint16_t buffer[5];

  // read MAC address register
  prism2_read_RID(PRISM2_RID_CNFOWNMACADDR, 0, buffer, 5);
  
  *mac++ = buffer[2];
  *mac++ = buffer[2]>>8;
  *mac++ = buffer[3];
  *mac++ = buffer[3]>>8;
  *mac++ = buffer[4];
  *mac++ = buffer[4]>>8;
}

#define PRISM2_MAXKEYLEN 13

void prism2_set_WEPkey(void) {
  uint16_t buffer[(PRISM2_MAXKEYLEN+3)/2];
  uint8_t keylen = (eeprom_read_byte(&config.flags) & CONFIG_KEY_WEP40)?5:13;
  
  printf_P(PSTR("Enabling %d bit WEP encryption\n"), 8*keylen);
  
  // prepare buffer for key write
  buffer[0] = (keylen+3)/2;
  buffer[1] = PRISM2_RID_CNFWEPDEFAULTKEY0;
  // copy key string to buffer
  eeprom_read_block((uint8_t*)&buffer[2], config.key, keylen);
  ((uint8_t*)&buffer[2])[keylen] = 0;

  // write default key 0
  prism2_write_RID(0, buffer);
  
  // activate first wep key
  prism2_write_RID2(PRISM2_RID_CNFWEPDEFAULTKEYID, 0);

  // set WEP active
  prism2_write_RID2(PRISM2_RID_CNFWEPFLAGS, 1);

  // configure authentication mode
  prism2_write_RID2(PRISM2_RID_CNFAUTHENTICATION, 1);
}

void prism2_init(void) {
  uint8_t result;
  
  // first init easily fails ...
  if( (result = prism2_command(PRISM2_CMD_INIT,0)) ) 
    if( (result = prism2_command(PRISM2_CMD_INIT,0)) ) 
      config_console_fatal(PSTR("Initialization failed"), 3);

  prism2_get_MAC(uip_ethaddr.addr);
  printf_P(PSTR("MAC: %02x:%02x:%02x:%02x:%02x:%02x\n"),
	   uip_ethaddr.addr[0], uip_ethaddr.addr[1], 
	   uip_ethaddr.addr[2], uip_ethaddr.addr[3], 
	   uip_ethaddr.addr[4], uip_ethaddr.addr[5]);

  printf_P(PSTR("Setting max pkt size: %d\n"), UIP_BUFSIZE+14);

  // set max packet size
  prism2_write_RID2(PRISM2_RID_CNFMAXDATALEN, UIP_BUFSIZE+14);
  
  // set operating mode / port type
  prism2_write_RID2(PRISM2_RID_CNFPORTTYPE, 
		    (eeprom_read_byte(&config.flags) & CONFIG_FLAG_MANAGED)?
		    PRISM2_MODE_BSS : PRISM2_MODE_IBSS);
  
  // set SSID
  prism2_set_SSID();

  // set channel
  prism2_write_RID2(PRISM2_RID_CNFOWNCHANNEL, 
		    eeprom_read_byte(&config.channel));

  // set key if encryption is enabled
  if(eeprom_read_byte(&config.flags) & (CONFIG_KEY_WEP40|CONFIG_KEY_WEP104))
    prism2_set_WEPkey();
  
  // enable the interface
  prism2_command(PRISM2_CMD_ENABLE_MAC0, 0);

  // request free buffer space to store outgoing frame
  prism2_command(PRISM2_CMD_ALLOC, UIP_BUFSIZE+14);

  // wait for buffer to be allocated
  while(!(cf_io_read_word(PRISM2_REG_EVSTAT) & PRISM2_EVENT_ALLOC) );

  // get the buffer FID
  txfid = cf_io_read_word(PRISM2_REG_ALLOCFID);
}


void prism2_event_check(void) {
  uint16_t evstat_data;

  evstat_data = cf_io_read_word(PRISM2_REG_EVSTAT);
	
  if(evstat_data & PRISM2_EVENT_TX) {
#ifdef DEBUG_EVENT
    printf_P(PSTR("Event: TX\n"));
#endif
    cf_io_write_word(PRISM2_REG_EVACK,PRISM2_EVENT_TX);
  }

  if(evstat_data & PRISM2_EVENT_TXEXEC) {
#ifdef DEBUG_EVENT
    printf_P(PSTR("Event: TXEXEC\n"));
#endif
    cf_io_write_word(PRISM2_REG_EVACK,PRISM2_EVENT_TXEXEC);
  }
  
  if(evstat_data & PRISM2_EVENT_ALLOC) {
#ifdef DEBUG_EVENT
    printf_P(PSTR("Event: ALLOC\n"));
#endif
    cf_io_write_word(PRISM2_REG_EVACK, PRISM2_EVENT_ALLOC);
  }
  
  if(evstat_data & PRISM2_EVENT_CMD) {
#ifdef DEBUG_EVENT
    printf_P(PSTR("Event: CMD\n"));
#endif
    cf_io_write_word(PRISM2_REG_EVACK, PRISM2_EVENT_CMD);
  }
  
  if(evstat_data & PRISM2_EVENT_INFO) {
#ifdef DEBUG_EVENT
    printf_P(PSTR("Event: INFO\n"));
#endif
    cf_io_read_word(PRISM2_REG_INFOFID);
    cf_io_write_word(PRISM2_REG_EVACK, PRISM2_EVENT_INFO);
  }
  
  if(evstat_data & PRISM2_EVENT_INFDROP) {
#ifdef DEBUG_EVENT
    printf_P(PSTR("Event: INFDROP\n"));
#endif
    cf_io_write_word(PRISM2_REG_EVACK, PRISM2_EVENT_INFDROP);
  }
  
  if(evstat_data & PRISM2_EVENT_WTERR) {
#ifdef DEBUG_EVENT
    printf_P(PSTR("Event: WTERR\n"));
#endif
    cf_io_write_word(PRISM2_REG_EVACK, PRISM2_EVENT_WTERR);
  }
}

uint16_t prism2_poll(uint8_t *packet) {
  uint16_t rxfid = 0;
  uint16_t packetLength = 0;

  prism2_event_check();

  // check if packets have been received
  if(cf_io_read_word(PRISM2_REG_EVSTAT) & PRISM2_EVENT_RX) {
    // we have a receive event
    // get RxFID
    rxfid = cf_io_read_word(PRISM2_REG_RXFID);
    // read the packet length
    prism2_read_BAP0(rxfid, 44, &packetLength, 1);
  }

  // if there's no packet or an error - exit
  if( !packetLength ) 
    return 0;
  
  // drop anything too big for the buffer
  if( packetLength+6 > UIP_BUFSIZE ) {
    // ACK the receive event to finish up
    cf_io_write_word(PRISM2_REG_EVACK, PRISM2_EVENT_RX);
    return 0;
  }
  
  // drive led via gpo0
  cf_set_gpo(CF_GPO0);

  // read ethernet addresses
  prism2_read_BAP0(rxfid, 2*23, (uint16_t*)&packet[0], 6);
  // and payload
  prism2_read_BAP0(rxfid, 2*33, (uint16_t*)&packet[12], (packetLength+6-1)>>1);
  // ACK the receive event to finish up
  cf_io_write_word(PRISM2_REG_EVACK, PRISM2_EVENT_RX);
  
  // all gpos low
  cf_set_gpo(0);

  return packetLength+6;
}

void prism2_setup_tx_header(uint16_t* header) {
  uint16_t i;
  
  // clear out header
  for(i=0;i<22;i++)
    header[i] = 0x00;

  // set TxRate and retry count
  header[5] = (0<<8) | 0;
  // 0x00 = automatic selection
  // 0x0A = 10  = 1.0Mbit/s
  // 0x14 = 20  = 2.0Mbit/s
  // 0x37 = 55  = 5.5Mbit/s
  // 0x6E = 110 = 11 Mbit/s
  
  // set TxControl
  header[6] = 0x0004;
  
  // write length
  // (not really needed since card will pull info from 802.3 header)
  //header[22] = len+8;
  
  // fill in 802.3 header fields
  header[30] = 0xAAAA;
  header[31] = 0x0003;
  header[32] = 0x0000;
  
  // src mac address @ byte offset 52
}

void prism2_send(uint8_t* packet, uint16_t len) {
  uint16_t TxHeader[34];
  uint16_t i;
  uint8_t stat;

  // drive led via gpo0
  cf_set_gpo(CF_GPO0);

  // adjust packet length because MAC addresses and type
  // will be written seperately from packet payload
  len -= 14;
  
  // write the outgoing frame to BAP
  // begin with control structure
  prism2_setup_tx_header(TxHeader);
  
  // write dest and src MAC addresses
  for(i=0;i<6;++i)
    TxHeader[23+i] = packet[i*2+1]<<8 | packet[i*2];

  // write length
  TxHeader[29] = htons(len+8);
  // write type
  TxHeader[33] = packet[13]<<8 | packet[12];
  
  // write Tx header out to BAP
  prism2_write_BAP0(txfid, 0, TxHeader, 34);
  // write packet out to BAP
  prism2_write_BAP0(txfid, 68, (uint16_t*)&packet[14], (len+1)>>1);
  // issue transmit command
  stat = prism2_command(PRISM2_CMD_TX_RECL, txfid);
  if(stat)
    printf_P(PSTR("Transmit failed: 0x%x\n"), stat);

  // all gpos low
  cf_set_gpo(0);
}

